/*
Sparse set of pointers scattered across terrain.

Orion Sky Lawlor, olawlor@acm.org, 2004/11/14
*/
#ifndef __OSL_TERRAIN_SPARSE_H
#define __OSL_TERRAIN_SPARSE_H

#include <vector>
#include "terrain/terrain.h"

/**
  Keeps a set of pointers associated with quads on the 
  Terrain.  Most of the pointers are assumed to be NULL.
*/
template <class T>
class TerrainSparse {
	/**
	  Stores a 2D array of T pointers, one for each
	  quad of one level of the tree.
	*/
	class Sparse2d {
		int side;
		std::vector<T *> data;
	public:
		Sparse2d(int side_) 
			:side(side_), data(side*side, (T *)0) { }
		~Sparse2d() {
			for (unsigned int i=0;i<data.size();i++)
				delete data[i];
		}
		T *&at(unsigned int x,unsigned int y) {
			return data[x+y*side];
		}
	};
	
	/**
	  data[tile][level]->at(x,y)
	*/
	std::vector< /* tile */
		std::vector< /* level */
			Sparse2d *
		>
	> data;
	
public:
	TerrainSparse(TerrainSet *terrain) 
		:data(terrain->totalTiles())
	{
		for (int tile=0;tile<terrain->totalTiles();tile++) {
			Terrain *t=terrain->getTile(tile);
			for (int level=0;level<=t->getMaxLevel();level++)
				data[tile].push_back(new Sparse2d(1<<level));
		}
	}
	~TerrainSparse() {
		for (unsigned int tile=0;tile<data.size();tile++)
			for (unsigned int level=0;level<data[tile].size();level++)
				delete data[tile][level];
	}
	
	/**
	  Apply this object to this quad.
	  Takes ownership of t, and will eventually delete or return it.
	  Returns the old thing if there was already something there.
	*/
	T *set(const TerrainAddress &addr,T *t) {	
		T *&ts=data[addr.tile][addr.level]->at(addr.x,addr.y);
		T *ret=ts;
		ts=t;
		return ret;
	}
	
	/**
	  Return the object associated with this quad, or 0 if none.
	*/
	T *get(const TerrainAddress &addr) {
		return data[addr.tile][addr.level]->at(addr.x,addr.y);
	}
};

#endif
