/*
Utility classes for guts of terrain simplification code.

Orion Sky Lawlor, olawlor@acm.org, 2003/12/6
*/
#ifndef __OSL_TERRAIN_UTILITY_H
#define __OSL_TERRAIN_UTILITY_H

#include "osl/vector3d.h"
using osl::Vector3d;
#include "osl/vector2d.h"
using osl::Point;
using osl::Vector2d;
#include "osl/viewpoint.h"

/// Bits for drawing mode.  These are chosen so the 
/// fastest, highest quality, and default drawing mode is 0.
enum {
	drawMode_frame=(1<<0), // outline triangles
	drawMode_number=(1<<1), // number triangle mode
	drawMode_backface=(1<<2), // don't cull backfaces
	drawMode_noclip=(1<<3), // don't do view clipping
	drawMode_notex=(1<<4), // don't draw texture
	drawMode_nomip=(1<<5), // don't do mipmapping
	drawMode_stats=(1<<6), // give statistics printouts
	drawMode_zbuf=(1<<7), // use the zbuffer
	drawMode_vieworder=(1<<8), // show the display order as color
	drawMode_noobjs=(1<<9), // don't draw objects decorating terrain 
	drawMode_crack=(1<<10), // use all quads, which results in cracks
	drawMode_last
};

class TerrainTexture; /* in render.h */
class TerrainObject; /* NEVER declared here; basically a void *.  Declare this class however you like, or not at all. */
class TerrainCamera;
class TerrainLevel;
class TerrainSet;

/// Identifies a TerrainObject inside the TerrainObjectManager
typedef int TerrainObjectID;

/// Tile-unique identifier for a list of TerrainObjects.
///  This is the identifier held by each terrain quad.
/// A listID of 0 means there are no objects in my quad.
typedef int TerrainObjectListID;


/**
  Addresses one individually drawable element, a quad,
  in the terrain.
*/
class TerrainAddress {
public:
	/// Tile number; Terrain within a TerrainSet
	///  (limited to 256x256 tiles)
	unsigned short tile;
	
	/// Bounding level within Terrain: 0 is 1x1 quads, 1 is 2x2, 2 is 4x4, etc.
	unsigned char level;
	
	/// X,Y address in our level's TerrainLevel image (quad/pixel number)
	///   This is our level's zero-based quad index.
	///   Our vertices are at (x,y), (x+1,y), (x,y+1), and (x+1,y+1).
	///   Our child quads are at level+1, 2*x+[0,1] by 2*y+[0,1].
	///  (limited to 65,536 pixels per side of bound image)
	unsigned short x,y;
	
	TerrainAddress() {}
	TerrainAddress(int tile_) :tile(tile_), level(0), x(0), y(0) {}

	/// Print address (for debugging)
	void print(void) const;
};

/**
Keeps track of objects pasted onto the terrain.  This is a separate class because:
 	- Sometimes you don't need to paste objects onto the terrain. (e.g., because you draw objects later yourself, or else because there *is* only terrain.)
	- Sometimes you've got a crazy complicated object system, like impostors.

So a single implementation can't make everybody happy.
The default implementation doesn't actually handle objects at all.
*/
class TerrainObjectManager {
public:
	virtual ~TerrainObjectManager();
	
	/**
	  Add this object to the terrain.
	  (used only during terrain setup)
	*/
	virtual void addObject(TerrainObject *obj);
	
	/**
	  Append this object to this quad's objectList.
	  \param addr Location of quad to add.
	  \param oldList Old object list, or 0 if no list existed already.
	  \param obj New object to add to list.
	  \param return New object list, containing all old objects plus obj.
	*/
	virtual TerrainObjectListID appendList(const TerrainAddress &addr,
		TerrainObjectListID oldList,TerrainObject *obj);
	
	/** Draw this list of terrain objects for a quad.
	  \param cam The terrain-rendering camera.
	  \param b The level of terrain for the quad being rendered.
	  \param addr The location in that terrain level for the quad being rendered.
	  \param objectList The listID for that quad's objects.
	*/
	virtual void drawList(const TerrainCamera *cam,const TerrainLevel *b,
		const TerrainAddress &addr,TerrainObjectListID objectList);
};

/**
 Draws everything related to a terrain.
 
 This allows us to traverse terrains in various
 different ways easily:
   - Regular drawing
   - Various outline/analysis drawing modes:
        - Wireframe  (FIXME: actually separate this version!) 
	- Drawing order
   - Render impostors (on parallel client)
   	- Impostors for ground plane
	- Impostors for objects/object lists
   - Decide which impostors to render (on parallel machine)
  	- Overrides terrainQuad and drawObjects.
*/
class TerrainRenderer {
protected:
	TerrainCamera *cam; ///< Camera currently being used.
public:
	TerrainRenderer() :cam(0) {}
	virtual ~TerrainRenderer();
	
	virtual void beginTerrain(TerrainCamera &cam,const TerrainSet *s);
	virtual void endTerrain(void);
	
	/// Starting to draw this tile.
	virtual void beginTile(const TerrainAddress &addr);
	/// Done drawing this tile.
	virtual void endTile(const TerrainAddress &addr);
	
	/**
	  Recursively expand the quad (x,y)-(x+2,y+2).
	  This can do arbitrary processing at each level of
	  the tree.  
	  Default implementation just calls b->expand(*cam,viewCull,x,y). 
	*/
	virtual void expand(const TerrainLevel *b,bool viewCull,int x,int y);
	
	/**
	  Draw one terminal (non-recursing, "leaf") 
	  quad of the ground.  Note that this should 
	  normally include a call to drawObjects, which
	  is all the default implementation does.
	*/
	virtual void drawQuad(const TerrainLevel *b,const TerrainAddress &addr);
	
	/**
	  Draw one triangle of the ground for this quad,
	  between these vertices.  The triangle is centered on 
	  this bound's vertex (addr.x+1,addr.y+1), and touches 
	  two other vertices, numbered clockwise from topleft:
	      (dx,dy)                vertex number v1 or v2
	      (0,0)  (1,0)  (2,0)      0  1  2
	      (0,1)  (1,1)  (2,1)      7  8  3
	      (0,2)  (1,2)  (2,2)      6  5  4 
	  
	  This renderer's drawObjects method will be called 
	  separately after drawing all needed triangles.
	*/
	virtual void drawTriangle(const TerrainLevel *b,const TerrainAddress &addr,int v1,int v2);
	
	/**
	  Draw all the objects for this terrain quad. 
	  list contains the quad's object list, which is 
	  guaranteed to be nonzero.  No other early-exit tests
	  are performed before the call, though, so check view!
	*/
	virtual void drawObjects(const TerrainLevel *b,const TerrainAddress &addr,TerrainObjectListID list);
};

/**
 Camera representation used by terrain system.
 This includes a camera location, orientation, 
 and error tolerance.
 
 Coordinate systems:
   Map coordinates: UTM projection coordinates (meters).
   World/OpenGL coordinates: centered on camera.
 
 This isn't 0,0,0 to avoid roundoff-induced camera shake,
 which for a UTM projection amounts to dozens of centimeters
 at middle latitudes when using OpenGL's float matrix ops.
*/
class TerrainCamera {
	/// Camera location-- center of projection in map coordinates.
	Vector3d loc;
	double invErr;
	int drawMode;
	Vector3d viewDir;
	
public:
// Planes used for view clipping:
	enum {nClip=4};
	osl::Halfspace3d clip[nClip];
	
	/// Destination for pieces of terrain coming out
	TerrainRenderer *renderer;
	
	/// Source for terrain objects
	TerrainObjectManager *objectMgr;
	
	/// Return if this map-coordinates bounding sphere is in the view volume.
	///   
	inline osl::boundview_t sphereInView(const Vector3d &center,double radius) const {
		osl::boundview_t ret=osl::boundview_all;
		for (int c=0;c<nClip;c++) {
			osl::boundview_t v=clip[c].hitSphere(center,radius);
			if (v==osl::boundview_none) return osl::boundview_none;
			if (v==osl::boundview_part) ret=osl::boundview_part;
		}
		return ret;
	}
	
//  (Set and used only in terrain.cpp)
	/// Terrain image pixel coordinates beneath the camera
	Point loc2d;
	/// If true, blending should be enabled
	bool wantBlend;
	/// Texture for terrain use (NULL if none)
	const TerrainTexture *terrainTex;
	
/// View statistics:
	enum {
		viewSides=16, // Indexed by "activeSides" bitvector for regular quads
		viewFinest=32, // Indexed by "activeSides" for finest-level quads
		viewInactive, // Inactive quads (!white.isActive())
		viewCullFalse, // Expanded quads that skipped view clip test
		viewCullTrue,  // Expanded quads that did the view clip test
		viewQuads, // Quad counter
		viewPolys, // Polygon counter (polygons not in quads)
		viewVertices, // Vertex counter (not in quads)
		viewLast};
	int viewFreq[viewLast];
#if 1 /* enable statistics collection */
	inline void view(int s) { viewFreq[s]++; }
	inline void view(int s,int n) { viewFreq[s]+=n; }
#else /* skip statistics collection */
	inline void view(int s) { }
	inline void view(int s,int n) { }
#endif

	/// Maps world coordinates to screen
	osl::Viewpoint vp;
	
	/**
	  Create new camera with these parameters.
	     \param vp Viewpoint, in world/OpenGL coordinates.
	*/
	TerrainCamera(const osl::Viewpoint &vp_,
		double pixelErrTol,int drawMode=0);
	~TerrainCamera();
	
	/// Return the camera's center of projection in map coordinates.
	inline const Vector3d &getLocation(void) const {return loc;}
	
	/// Return the camera's approximate map view direction (NOT unit)
	inline const Vector3d &getViewDirection(void) const {return viewDir;}
	
	/// Return the camera's center in map coordinates (meters)
	inline osl::Vector2d getMap(void) const {return osl::Vector2d(loc.x,loc.y);}
	
	/// Return the inverse of the camera's maximum
	///   projected error tolerance, in radians.
	inline double getInvError(void) const {return invErr;}
	
	inline int getDrawMode(void) const {return drawMode;}
	
	void viewStats(void);
};



#endif
