/*
Terrain heirarchical data structure setup code.

This is my own take on Peter Lindstrom and Valerio Pascucci's 
simplification algorithms.

Orion Sky Lawlor, olawlor@acm.org, 2003/12/6
*/
#include "terrain/terrain.h"
#include "osl/least_squares.h"
using namespace osl;
using namespace osl::ru;

/*********** Terrain Objects ***********/
/// Add this object to our terrain
void TerrainSet::add(TerrainObjectManager *mgr,TerrainObject *obj,const osl::Bbox2d &box) 
{
	mgr->addObject(obj);
	// FIXME: use efficient algorithm based on bounding box and toTiles
	for (unsigned int i=0;i<t.size();i++) 
	if (t[i]) {
		if (t[i]->getGeo().getBox().intersects(box))
			t[i]->add(mgr,obj,box);
	}
}

/// Add this object to our terrain
void Terrain::add(TerrainObjectManager *mgr,TerrainObject *obj,const osl::Bbox2d &box) {
	for (int l=0;l<=nLevels;l++) {
		b[l]->add(mgr,obj,box);
	}
}

/// Add this object to our terrain-- adds it to every quad
void TerrainLevel::add(TerrainObjectManager *mgr,TerrainObject *obj,const osl::Bbox2d &box) {
	int x,y;
	OSL_RECT_LOOP(geo.pixelRectFmMap(box),x,y) 
	{
		TerrainQuad &q=quad.write(x,y);
		TerrainAddress addr=myAddr;
		addr.x=x; addr.y=y;
		q.objectList=mgr->appendList(addr,q.objectList,obj);
	}
}


/********** TerrainBound ****************/
void TerrainBound::merge(const Vector3d &V,
           const TerrainBound &child,const Vector3d &childV)
{
	if (child.error>error) error=child.error;
	// error=0.0;
	double child_r=V.dist(childV)+child.r;
	if (child_r>r) r=child_r;
}

/********** TerrainLevel: bound setup **************/

Terrain::Terrain(const osl::GeoHeight *ht_)
	:ht(ht_), heightSource(0), tex(0)
{
/* Build the levels of the bounding heirarchy */
	nLevels=size2level(ht->width);
	setupLevels(true);
}

// Create and initialize levels 1..nLevels
void Terrain::setupLevels(bool propagate) {
	int i;
	nLevels=size2level(ht->width);
	for (i=0;i<=nLevels;i++) {
		TerrainAddress boundAddr=myAddr;
		boundAddr.level=i;
		b[i]=new TerrainLevel(level2size(i),ht->width,boundAddr);
	}
	b[nLevels+1]=0; // for b[i+1] below, and for paranoia.
	for (i=0;i<=nLevels;i++) 
		b[i]->setup(ht,b[i+1],(i==0)?0:b[i-1]);
	
	if (propagate) {
		for (i=0;i<=nLevels;i++) b[i]->allocate();
	
		/// Link the levels from fine to coarse:
		for (i=nLevels;i>=0;i--) {
			double startTime=oglTimer();
			b[i]->propagate((const osl::GeoHeight *)ht);
			printf(" created bounding level %d: %.3f s\n",i,oglTimer()-startTime);
		}
	}
}
Terrain::~Terrain() {
	for (int i=0;i<=nLevels;i++)
		delete b[i];
	delete heightSource;
	delete tex;
}

TerrainLevel::TerrainLevel(int my_sz,int full_sz,TerrainAddress myAddr_)
	:sz(my_sz), stride((full_sz-1)/(my_sz-1)), myAddr(myAddr_),
	 bound(0, sz,sz), alloc_bound(0), quad(0,sz-1,sz-1), alloc_quad(0),
	 finer(0)
	// scanConv(osl::Rect(0,0,stride*(sz-1)+1,stride*(sz-1)+1))
{
	
}
TerrainLevel::~TerrainLevel() {
	delete[] alloc_bound;
	delete[] alloc_quad;
}
void TerrainLevel::setup(const osl::GeoImage *ht,
	const TerrainLevel *finer_,const TerrainLevel *coarser_)
{
	geo=osl::GeoImage(ht->origin,ht->pixelSize*stride,sz-1,sz-1);
	
	finer=finer_;
	coarser=coarser_;
	// printf("size: %f,%f/ origin: %f,%f\n",pixelSize.x,pixelSize.y, origin.x,origin.y);
}

void TerrainLevel::allocate(void)
{
	alloc_bound=new TerrainBound[sz*sz];
	bound.setData(alloc_bound);
	alloc_quad=new TerrainQuad[(sz-1)*(sz-1)];
	quad.setData(alloc_quad);
}
void TerrainLevel::propagate(const osl::GeoHeight *ht) {
	boundHt(ht);
	if (finer) boundFinerCenter();
	boundSide();
}


/**
  Propagate the approximation errors from the center vertices
  of the finer level of the approximation up to our pixels:
	D     D
	   S
	D     D
*/
void TerrainLevel::boundFinerCenter(void) {
	// Loop over all of us, except outer border row and column
	for (int y=0;y<sz-1;y++)
	for (int x=0;x<sz-1;x++) {
		// (srcX,srcY) is white vertex on finer grid
		int srcX=x*2+1, srcY=y*2+1;
		for (int dy=0;dy<2;dy++)
		for (int dx=0;dx<2;dx++) 
			merge(x+dx,y+dy, *finer, srcX,srcY);
	}
}
	
/**
  Propagate the approximation errors from our side vertices
  to our center vertices.
	  S
       S  D  S
	  S
*/
void TerrainLevel::boundSide(void) {
	for (int y=0;y<sz;y++)
	for (int x=0;x<sz;x++) {
		switch (vertex_kind(x,y)) {
		case vertex_corner: 
		case vertex_hside:
		case vertex_vside:
			continue;
		case vertex_center:
			merge(x,y, *this,x-1,y  );
			merge(x,y, *this,x  ,y+1);
			merge(x,y, *this,x+1,y  );
			merge(x,y, *this,x  ,y-1);
			break;
		};
	}
}

/**
  Compute the approximation errors at *this* level of the 
  heirarchy-- the height difference that would be present
  if each center or side vertex isn't activated.
*/
void TerrainLevel::boundHt(const osl::GeoHeight *ht) 
{
	int x,y;
	// Set Z values for our vertices (so quadError, below, will work)
	for (y=0;y<sz;y++)
	for (x=0;x<sz;x++) {
		bound.write(x,y).setHeight(ht->at(x*stride,y*stride));
	}
	
	// Compute errors and radii for our vertices
	for (y=0;y<sz;y++)
	for (x=0;x<sz;x++) {
		double err_z=0; // Maximum error
		double targ_r=0; // target radius
		switch (vertex_kind(x,y)) {
		case vertex_corner: 
			continue;
		case vertex_hside:
			err_z=quadError(ht,
				Point2d(x,y+1),
					Point2d(x+1,y),
			  Point2d(x-1,y),
				Point2d(x,y-1), true
			);
			targ_r=getV(x,y).dist(getV(x,y-1));
			break;
		case vertex_vside:
			err_z=quadError(ht,
				Point2d(x,y+1),
				         Point2d(x+1,y),
			  Point2d(x-1,y),
				Point2d(x,y-1), false
			);
			targ_r=getV(x,y).dist(getV(x-1,y));
			break;
		case vertex_center: {
			err_z=0; // Error will be set by coarser level's quad
			targ_r=getV(x,y).dist(getV(x-1,y-1));
			} break;
		};
		bound.write(x,y).setRadius(targ_r);
		bound.write(x,y).setError(err_z);
	}
	
	if (1)
	// Compute errors on our quads
	for (y=0;y<sz-1;y++)
	for (x=0;x<sz-1;x++) {
		double err_z=0; // Maximum error
		
		// There are two ways to split the quad (x,y): pick the best
		unsigned short flags=0;
		double err_diag=quadError(ht, /* split along x,y to x+1,y+1 */
			Point2d(x,y+1), Point2d(x+1,y+1),
			Point2d(x,y  ), Point2d(x+1,y  ),  false
		);
		double err_crs=quadError(ht, /* split along x,y+1 to x+1,y */
			Point2d(x,y+1), Point2d(x+1,y+1),
			Point2d(x,y  ), Point2d(x+1,y  ),  true
		);
		bool use_diag=err_diag<err_crs; /* use smaller error */
		if (0) use_diag=(myAddr.level&1); /* Original Lindstrom method: alternate sides */
		
		if (use_diag) {flags|=1; err_z=err_diag;}
		else /* err_false smaller*/ {/* flags|=0 */ err_z=err_crs;}
		
		quad.write(x,y).flags=flags;
		if (finer) {
			TerrainLevel *f=(TerrainLevel *)finer;
			f->bound.write(x*2+1,y*2+1).setError(err_z);
		}
	}
}


/// Return the total error from using this quad:
///   if true, split the a-d axis as triangle(abd)/triangle(dca);
///   if false, split the b-c axis as triangle(abc)/triangle(bdc).
/*
4 options here:

 (split)   (!split)
  
 A \  B     A  / B    rectangular
 C  \ D     C /  D

    A	       A      diamond
 C  |  B    C --- B
    D	       D
*/
double TerrainLevel::quadError(const GeoHeight *ht,
	const Point2d &a,const Point2d &b,
	const Point2d &c,const Point2d &d,
	bool split)
{
	double maxErr=0;
	if (split) { /* split a-d axis */
	   triError(maxErr,ht,a,b,d); triError(maxErr,ht,a,c,d);
	} else { /* split b-c axis */
	   triError(maxErr,ht,a,b,c); triError(maxErr,ht,b,c,d);
	}
	return maxErr;
}

/**
 Linear 2d position-to-height function:
 	h=a+b*x+c*y;
*/
class TerrainLinHeightFn {
	double a,b,c;
	void checkErr(const Point2d &p,double ht) {
		double tol=1.0e-3;
		double err=at(p)-ht;
		if (fabs(err)>tol) bad("heightFn matrix incorrect!");
	}
public:
	TerrainLinHeightFn(double h1,const Point2d &p1,
		double h2,const Point2d &p2,
		double h3,const Point2d &p3)
	{
		/* solve
		    [ 1 p1.x p1.y ] [ a ]   [ h1 ]
		    [ 1 p2.x p2.y ] [ b ] = [ h2 ]
		    [ 1 p3.x p3.y ] [ c ]   [ h3 ]
		*/
		MatrixT<double,3,4> m;
		m(0,0)=1.0; m(0,1)=p1.x; m(0,2)=p1.y; m(0,3)=h1;
		m(1,0)=1.0; m(1,1)=p2.x; m(1,2)=p2.y; m(1,3)=h2;
		m(2,0)=1.0; m(2,1)=p3.x; m(2,2)=p3.y; m(2,3)=h3;
		if (!m.solve()) bad("heightFn matrix is singular!");
		a=m(0,3);
		b=m(1,3);
		c=m(2,3);
		checkErr(p1,h1); checkErr(p2,h2); checkErr(p3,h3); 
	}
	
	/// Return the height at this point
	inline double at(int x,int y) const {return a+b*x+c*y;}
	inline double at(const Point2d &p) const {return at(p.x,p.y);}
};

static inline Vector2d p2v(const Point2d &p) {
	return Vector2d(p.x+0.5,p.y+0.5);
}

static inline Point2d mid(const Point2d &a,const Point2d &b) {
	if (a.distSqr(b)<2*2) return a;
	return Point2d((a.x+b.x)>>1,(a.y+b.y)>>1);
}

static inline void addError(double &maxErr,const TerrainLinHeightFn &h,const GeoHeight* ht, const Point2d &p)
{
	// printf("(%d,%d) ",p.x,p.y);
	double e=fabs(h.at(p.x,p.y)-ht->at(p.x,p.y));
	if (e>maxErr) maxErr=e;
}

/**
  Return the error of this tiny triangle.
  Tiny triangles have no interior points, so we estimate
  the error by extrapolation to nearby exterior points.
  Note: this makes the finest-level terrain look better, by
  splitting quads along the best axis.
*/
static void tinyTriError(double &maxErr,
	const TerrainLinHeightFn &h,const GeoHeight *ht,
	const Point2d &a,const Point2d &b,const Point2d &c)
{
	/* return; // ignore tiny triangles, which causes finest level to never be displayed */
	addError(maxErr,h,ht,ht->clipPixel(a+(a-b)));
	addError(maxErr,h,ht,ht->clipPixel(a+(a-c)));
	addError(maxErr,h,ht,ht->clipPixel(b+(b-a)));
	addError(maxErr,h,ht,ht->clipPixel(c+(c-a)));
}

/// Return the total error, computed recursively, from this triangle:
///  The points are given at the *finest* resolution.
static void triErrorRecursive(double &maxErr,
	const TerrainLinHeightFn &h,const GeoHeight *ht,
	const Point2d &a,const Point2d &b,const Point2d &c)
{
	/*
	        a
	      ab  ac
	    b   bc   c
	*/
	Point2d ab=mid(a,b), bc=mid(b,c), ac=mid(a,c);
	bool abe=(ab!=a && ab!=b);
	bool bce=(bc!=b && bc!=c);
	bool ace=(ac!=a && ac!=c);
	if ((!abe) && (!bce) && (!ace)) tinyTriError(maxErr,h,ht,a,b,c);
	if (!(abe || bce || ace)) return; /* degenerate terminating triangle */
	if (abe) addError(maxErr,h,ht,ab);
	if (bce) addError(maxErr,h,ht,bc);
	if (ace) addError(maxErr,h,ht,ac);
	if (abe && bce) triErrorRecursive(maxErr,h,ht,ab,bc,b);
	if (abe && ace) triErrorRecursive(maxErr,h,ht,ab,ac,a);
	if (bce && ace) triErrorRecursive(maxErr,h,ht,bc,ac,c);
	if (abe && bce && ace) triErrorRecursive(maxErr,h,ht,ab,bc,ac);
}

#if 0 /* osl scan conversion */
#include "osl/rasterizer_util.h"
using osl::ru::ScanConverted;

// Scan conversion tool, for use during quadError routine.
//  FIXME: gotta initialize this with whole-terrain rectangle...
static ScanConverted scanConv;

/// Compute the approximation error for this triangle 
///  by scan converting (rasterizing) the set of affected points.
void triErrorScan(double &maxErr,
		const TerrainLinHeightFn &h,const GeoHeight *ht,
		const Point2d &a,const Point2d &b,const Point2d &c)
{
	scanConv.resetAdd();
	osl::ru::PolyEdgeSimple pe(scanConv);
	Vector2d ao=p2v(a), bo=p2v(b), co=p2v(c);
	if (1) { /* enlarge the region slightly, to avoid pixel center roundoff */
		ao+=0.0001*(ao-0.5*(bo+co));
		bo+=0.0001*(bo-0.5*(co+ao));
		co+=0.0001*(co-0.5*(ao+bo));
	}
	pe.line(ao,bo); 
	pe.line(bo,co); 
	pe.line(co,ao);
	scanConv.prepareEnterExit(0);
	
	int maxHeight=ht->height;
	for (int y=0;;y++)
	{
	   ScanLine row(scanConv,y);
	   if (y>=maxHeight) break;
	   
	   for (int i=0;i<row.spans();i++) 
	   if (row[i].alpha!=0) {
	   	int xmin=row[i].x, xmax=row[i+1].x;
		for (int x=xmin;x<xmax;x++) 
		  addError(maxErr,h,ht,Point2d(x,y));
	   }
	}
}
#endif /* end osl scanline stuff */

/** Lets you look up the intersection of constant-Y scanlines with another line */
class scanline_slope {
	/*  x = y*m+o  */
	double d,m,o;
public:
	/** Compute the slope of the line between these two points */
	scanline_slope(const Point2d &a,const Point2d &b) {
		d=a.y-b.y;
		m=(a.x-b.x)/d;
		o=a.x-a.y*m; /* rearrange x = y*m+o */
	}
	/// Return true if this line is parallel to the scanlines
	bool is_degenerate(void) const {return d==0.0;}
	/// Return the intersection of our line with this scanline
	double get_x(int y) const {return y*m+o;}
};
/// Return the total error, computed using a loop, from approximating this terrain with a single triangle.
void TerrainLevel::triErrorLoop(double &maxErr,
		const TerrainLinHeightFn &h,const GeoHeight *ht,
		const Point2d &a,const Point2d &b,const Point2d &c)
{
	if (!(c.y<a.y)) osl::bad("triError called on invalid points: must c.y<a.y");
	if (!(c.y<=b.y)) osl::bad("triError called on invalid points: must c.y<=b.y");
	if (!(b.y<=a.y)) osl::bad("triError called on invalid points: must b.y<=a.y");
	
	// Fit linear function-of-y slope functions: x = m * y + o
	scanline_slope ab(a,b), bc(b,c), ca(c,a);
	/* ca is never degenerate */
	int x,y;
	for (y=c.y; y<=b.y; y++) {
		int left,right;
		left=(int)(bc.get_x(y));
		if (bc.is_degenerate()) { /* b.y==c.y */
			left=b.x;
		}
		right=(int)(ca.get_x(y));
		if (left>right) std::swap(left,right);
		for (x=left;x<=right;x++)
			addError(maxErr,h,ht,Point2d(x,y));
	}
	for ( /*y=b.y+1*/ ; y<=a.y; y++) {
		int left,right;
		left=(int)(ab.get_x(y)); /* ab is never degenerate if loop executes */
		right=(int)(ca.get_x(y));
		if (left>right) std::swap(left,right);
		for (x=left;x<=right;x++)
			addError(maxErr,h,ht,Point2d(x,y));
	}
}

/// Return the total height error from using this triangle:
/**
  In general, this could be totally horrible.  But luckily, the points abc
are in a specific arrangment, since this routine is only ever called by 
quadError.  Specifically, c.y <= b.y <= a.y, and c.y<a.y.
*/
void TerrainLevel::triError(double &maxErr,const GeoHeight *ht,Point2d a,Point2d b,Point2d c)
{
	// Don't mess with off-screen triangles (e.g., from black vertices)
	if (clip(a) || clip(b) || clip(c)) return;
	
	// Expand to full resolution:
	a*=stride; b*=stride; c*=stride;
	
	// Fit linear height function
	TerrainLinHeightFn h(
		ht->at(a.x,a.y),a,
		ht->at(b.x,b.y),b,
		ht->at(c.x,c.y),c
	);

#if 1
/* use fastest method */
	// Recursively break down small triangles (faster than looping, and more accurate too):
	const double recurseTol=3.0;
	const int recurseTolSqr=(int)(recurseTol*recurseTol);
	if (a.distSqr(b)<recurseTolSqr) //Recursively traverse small triangles
		triErrorRecursive(maxErr,h,ht,a,b,c);
	else // Rasterize large triangles:
		triErrorLoop(maxErr,h,ht,a,b,c);
#else
/* use all methods, and compare the results (be sure to skip triErrorTiny!): */
	double rErr=0, sErr=0, lErr=0;
	triErrorRecursive(rErr,h,ht,a,b,c);
	triErrorScan(sErr,h,ht,a,b,c);
	triErrorLoop(lErr,h,ht,a,b,c);
	if (fabs(rErr-sErr)>0.01 ||  fabs(sErr-lErr)>0.01) {
		breakhere(); // re-run, in debugger:
		triErrorRecursive(rErr,h,ht,a,b,c);
		triErrorScan(sErr,h,ht,a,b,c);
	 	printf("\nBad error computation for tri (%d,%d) (%d,%d) (%d,%d) -> errors %.2f, %.2f, %.2f\n",
	 		a.x,a.y, b.x,b.y, c.x,c.y, rErr,sErr,lErr);
		abort();
	}
	if (maxErr<rErr) maxErr=rErr;
#endif
}



